/* word-wrapping HyperCard XFCN that returns a chosen number 'lines' of
 * text, each line <= 'maxLineLength', with a word beginning at file
 * location 'target' on line 'targetLine' (counting from line #1),
 * from a chosen text 'file' ... called as:
 *		wrapText (lines, target, targetLine, maxLineLength, file)
 *
 * The results are returned as a nice zero-terminated string that can
 * be stuffed into a text field as desired.  This XFCN is designed
 * to be useful in retrieving contents for the full-text card of
 * Texas.  It has some complexity to handle word-wrapping nicely....
 *
 * The XFCN also returns some additional information in the two lines
 * after the last line returned:
 * firstTextByte (the byte number of the first character in the first line
 *					of the returned string), and
 * nextTextByte (the byte number of the next character AFTER the last line
 *					of the returned string).
 * (The values don't count any added '\r' return characters used to
 * make line breaks.)  They should be useful in scrolling or paging
 * the text display window up and down....
 *
 * Thus, for example, to grab 19 lines of text with the word starting
 * at byte number 12345 placed on the middle line (line number 10 of the
 * set 1 through 19), with maximum line length of 80 characters,
 * from file refNum 666, call:
 *		wrapText (19, 12345, 10, 80, 666)
 * To do a scroll-down of one line from there, call:
 *		wrapText (19, nextTextByte, 19, 80, 666)
 * To do a page-up from there, call:
 *		wrapText (19, firstTextByte - 1, 19, 80, 666)
 * etc.
 *
 * A limitation:  the value of 'targetLine' MUST be in the range 1 through
 * 'lines', inclusive ... maybe remove this restriction someday, but not
 * yet!
 *
 * If some bytes are required from outside the actual file boundaries
 * (from a byte number <0 or >EOF) to fill out the request number of
 * lines, empty lines are returned for the 'unreal' areas.  If,
 * however, a target outside the file boundaries is requested, the
 * function refuses, beeps, and returns a null string....as it does
 * for any detected I/O error, etc.
 *
 * All bytes to be returned are checked and any '\0' bytes are
 * turned into spaces, to be safe and avoid problems with HyperCard
 * mistaking the '\0' for an end-of-string delimiter....
 *
 * Function is stored as XFCN number 538, named "wrapText"....
 *
 * 871129, 871208-21- ... ^z
 */

#include <MacTypes.h>
#include <FileMgr.h>
#include <HyperXCmd.h>

#define LONGEST_LINE 127
#define MOST_LINES    63
#define EOF			 (-1)

pascal void main (XCmdBlockPtr paramPtr);
long getTheText (int lines, long target, int targetLine, int maxLineLength,
	int refNum, char *ans, XCmdBlockPtr paramPtr);
char *buildNextLine (int maxLineLength, char *textp, char *outp);
void complain (XCmdBlockPtr paramPtr);
long atol (char *s);
int putNum (char *ans, long count);
int copyLine (char *cp, char *ans);

pascal void main (paramPtr)
  XCmdBlockPtr paramPtr;
  {
	int refNum, lines, targetLine, maxLineLength;
	long target, answer_bytes, maxTarget;
	Handle answer;
	
	if (paramPtr->paramCount != 5)
	  {
	  	complain (paramPtr);
	  	return;
	  }
	 
	lines = atol (*(paramPtr->params[0]));
	target = atol (*(paramPtr->params[1]));
	targetLine = atol (*(paramPtr->params[2]));
	maxLineLength = atol (*(paramPtr->params[3]));
	refNum = atol (*(paramPtr->params[4]));
	GetEOF (refNum, &maxTarget);

	if (lines < 1 || lines > MOST_LINES || target < 0 ||
			target > maxTarget || targetLine < 1 || targetLine > lines ||
			 maxLineLength < 1 || maxLineLength > LONGEST_LINE ||
			 refNum == 0)
	  {
	  	complain (paramPtr);
	  	return;
	  }
		
	answer = NewHandle (lines * (maxLineLength + 1) + 32);
	HLock (answer);
	answer_bytes = getTheText (lines,  target, targetLine, maxLineLength,
				refNum, *answer, paramPtr);
	*(*answer + answer_bytes) = '\0';
	paramPtr->returnValue = answer;
	HUnlock (answer);
	return;
  }


/* function to fetch the requested chunk of text from the file ... handle
 * problems with out-of-bounds requests (turn the characters not fetched
 * into '\r', empty lines) and with '\0' characters in text file (turn
 * them into ' ', spaces) ... and let HyperCard know the proper values
 * of firstTextByte and nextTextByte global (Texas stack) variables by
 * putting them in lines of their own at the very end of the returned
 * string.  Return the total number of characters put into answer string
 * as the result of this function.
 *
 * Strategy:  begin at byte number 'target' in the source file.  From that
 * point, grab a chunk of the text file guaranteed to be more than
 * big enough to hold the range of lines required in the worst case:
 * (lines + 2) * (maxLineLength + 1) bytes, ranging
 * from      target - (targetLine + 1) * (maxLineLength + 1)
 * through   target + (lines - targetLine + 1) * (maxLineLength + 1).
 * Grab the extra text at the start in order to make word-wrapping and
 * scrolling be smoother in the most common cases....
 *
 * (It is probably more efficient to do this than to spend time computing
 * a sharper bound on the bytes to read from disk, based on scanning '\r's,
 * etc.  The chunk will never be more than about 8kB long, and in a
 * typical case of 80 columns and 20 rows, it's only about 1.6kB.)
 * Handle any worries about being near the beginning of the
 * text file by putting extra '\r's at start of buffer; provide a
 * safety net against pathological cases (files with thousands of
 * spaces and no returns, etc.) by putting plenty of '\r's at end of
 * buffer.
 *
 * Now, begin at the beginning of that chunk of text and call the function
 * buildNextLine() repeatedly.  That function handles the magic
 * and mystery of word-wrapping, and builds lines of text
 * in a chosen location, an array linesArray.  Let it use
 * up the text chunk, cycling through the lines in linesArray
 * until the line containing byte number 'target' has been reached.
 * Then go on from there for another 'lines'-'targetLine' lines of text.
 * The result is the information needed for the returned result to
 * HyperCard.  Copy it into the answer string and finish up by putting
 * the values of firstTextByte and nextTextByte as two lines of text
 * strings at the end....
 *
 * The function buildNextLine() does all the dirty work of
 * figuring out where a line ends, putting the right characters
 * into the array, and terminating the line with a '\r' properly.  Note
 * that the counting of characters to maxLineLength does NOT include
 * the '\r'.  Note that the value returned by buildNextLine() is the
 * ACTUAL pointer to the final character used from the text buffer.
 * Note that linesArray is big enough to hold an extra '\r'.  The
 * array lineStart holds the byte offset for the start of each line in
 * linesArray, relative to the beginning of the text file.
 *
 * The function copyLine() copies a line from the linesArray to the
 * output string, and returns the number of actual characters copied,
 * including the terminal '\r' for each line. It also filters out any
 * '\0' character, to avoid problems with HyperCard mistaking that for
 * the end of the answer string....
 */

long getTheText (lines, target, targetLine, maxLineLength, refNum, ans,
		paramPtr)
  long target;
  int lines, targetLine, maxLineLength, refNum;
  char *ans;
  XCmdBlockPtr paramPtr;
  {
	char linesArray[MOST_LINES][LONGEST_LINE+1], *textp, *new_textp, *ans0,
			text[(MOST_LINES+2)*(LONGEST_LINE+1) + 2*MOST_LINES];
	long firstTextByte, nextTextByte, bytes, lineStart[MOST_LINES],
			offset;
	int i, current_line, targetArrayLine;
	
	ans0 = ans;
	textp = text;
	offset = target - (targetLine + 1) * (maxLineLength + 1);
	bytes = (lines + 2) * (maxLineLength + 1);
	*textp++ = '\r';						/* put some '\r's at start */
	if (offset < 0)
	  {
		for (i = 0; i < targetLine; ++i)
			*textp++ = '\r';
		offset = 0;
	  }

	SetFPos (refNum, fsFromStart, offset);
	i = FSRead (refNum, &bytes, textp);
	if (i != noErr && i != eofErr)
	  {
	  	SysBeep (10);
		return (0);					/* give up on bad disk read error! */
	  }
	for (i = 0; i < lines; ++i)
		*(textp + bytes + i) = '\r';	/* put safety net '/r's at end */
	
	offset -= textp - text - 1;
	current_line = -1;
	textp = text + 1;
	do
	  {
		++current_line;
		if (current_line == MOST_LINES)
			current_line = 0;
		lineStart[current_line] = offset;
		new_textp = buildNextLine (maxLineLength, textp,
					linesArray[current_line]);
		offset += new_textp - textp;
		textp = new_textp;
	  }
	while (offset <= target);
	
	targetArrayLine = current_line;
	for (i = targetLine + 1; i <= lines; ++i)
	  {
		++current_line;
		if (current_line == MOST_LINES)
			current_line = 0;
		lineStart[current_line] = offset;
		new_textp = buildNextLine (maxLineLength, textp,
					linesArray[current_line]);
		offset += new_textp - textp;
		textp = new_textp;
	  }
	nextTextByte = offset;
	
	current_line = targetArrayLine - targetLine + 1;
	if (current_line < 0)
		current_line += MOST_LINES;
	firstTextByte = lineStart[current_line];
	for (i = 0; i < lines; ++i)
	  {
		ans += copyLine (linesArray[current_line], ans);
		++current_line;
		if (current_line == MOST_LINES)
			current_line = 0;
	  }
	
	ans += putNum (ans, firstTextByte);
	*ans++ = '\r';
	ans += putNum (ans, nextTextByte);
	*ans++ = '\r';
	
	return (ans - ans0);
  }


/* function to beep and set the return string to null (= "")
 */

void complain (paramPtr)
  XCmdBlockPtr paramPtr;
  {
	Handle answer;
	
  	SysBeep (10);
	answer = NewHandle (1);
	**answer = '\0';
	paramPtr->returnValue = answer;
	return;
  }


/* function to convert alphabetic string to a long integer ... from LSC
 * library.... simplified to avoid using isspace() & isdigit() ....
 */

long atol (s)
  register char *s;
  {
	register char signflag = 0;
	register long r = 0;

	while ((*s == ' '))
		s++;
		
	if (*s == '-')
	  {
		signflag = 1;
		s++;
	  }
	else if (*s == '+')
	 	s++;

	while (*s >= '0' && *s <= '9') 
		r = r * 10 + (*s++ - '0');
	
	return (signflag ? -r : r);
}


/* function to convert a number into a string and put it into the chosen
 * target place ... returns the number of characters stored ...
 * based on K&R p. 60 example of itoa().... ^z
 */

int putNum (ans, num)
  register char *ans;
  register long num;
  {
	register int i, j;
	int s, result;
	
	i = 0;
	s = 1;
	if (num < 0)
	  {
		num = -num;
		s = -1;
	  }
	
	do
	  	ans[i++] = num % 10 + '0';
	while ((num /= 10) > 0);
	
	if (s < 0)
		ans[i++] = '-';
	result = i;
	
	for (--i, j = 0; j < i; ++j, --i)
	  {
	  	s = ans[i];
	  	ans[i] = ans[j];
	  	ans[j] = s;
	  }

	return (result);
  }


/* function to build a line of text, handling word-wrapping nicely; takes
 * in the value of maxLineLength, a pointer to where characters from the
 * text file are to be found (textp), and a pointer (outp) to where the
 * formatted line is to be placed.  This is where all the magic resides.
 *
 * The method is:
 *  - if previous character to the initial *textp is not a '\r',
 *		scan in characters until encountering a non-space (that is,
 *		skip over embedded spaces);
 *  - put characters into the output string and count them until reaching
 *		a '\r', or until a total of maxLineLength characters has been
 *		placed in the output string;
 *	- if the final character was a '\r', put it out;
 *	- else the maxLineLength ceiling has been hit, so back up starting at
 *	 	the last character put out, until a space character
 *		is encountered or until all the characters in this line
 *		have been examined.  If there were no space characters, put a
 *		'\r' at the very end of the line.  Otherwise, continue
 *		backing up until a non-space is found or until all the characters
 *		in this line have been examined, at which point, put an artificial
 *		'\r' after that point;
 *	- return the final value of textp ... so that the calling routine
 *		can tell how many characters have really been used from the text
 *		file.
 *
 * Since this function is applied to an input buffer which is GUARANTEED
 * to have a terminating '\r', there is no worry about running off the
 * end of the buffer....
 */

char *buildNextLine (maxLineLength, textp, outp)
  int maxLineLength;
  register char *textp, *outp;
  {
	char *textp0;

	if (*(textp - 1) != '\r')
		while (*textp == ' ')
			++textp;

	textp0 = textp;

	while (*textp != '\r' && (textp - textp0) < maxLineLength)
		*outp++ = *textp++;

	if (*textp == '\r')
		*outp = *textp++;
	else					 /* here textp == textp0 + maxLineLength */
	  {
		if (*textp != ' ')
			while (*--textp != ' ' && textp > textp0) 
				;						 /* drop final word fragment */
		if (textp == textp0)
		  {								/* line was solid characters */
			textp += maxLineLength;
			*outp = '\r';
		  }
		else
		  {
			while (*--textp == ' ' && textp > textp0)
				;							 /* drop trailing spaces */
			if (textp == textp0)
			  {								/* line was solid spaces */
				textp += maxLineLength;
				*outp = '\r';
			  }
			else
			  {
				++textp;
				outp -= textp0 - textp + maxLineLength;
				*outp = '\r';
			  }
		  }
	  }
	
	return (textp);
  }


/* function to copy a line terminated by '\r' from one place to
 * another, returning the number of characters copied....also filters
 * any '\0' characters to spaces....
 */

int copyLine (cp, ans)
  register char *cp, *ans;
  {
	char *cp0;
	
	cp0 = cp;

	while (*cp != '\r')
		*ans++ = (*cp ? *cp++ : (cp++ , ' '));

	*ans = *cp++;

	return (cp - cp0);
  }




